from contextlib import asynccontextmanager
from subprocess import PIPE

from tornado.testing import gen_test
from traitlets.config.loader import Config

try:
    from .. import (
        LimitedKernelManager,
        MaximumKernelsException,
    )
except ImportError:
    pass

from .utils import TestAsyncKernelManager


# Test that it works as normal with default config
class TestLimitedKernelManager(TestAsyncKernelManager):
    __test__ = True

    # static so picklable for multiprocessing on Windows
    @staticmethod
    @asynccontextmanager
    async def _get_tcp_km():
        c = Config()
        km = LimitedKernelManager(config=c)
        try:
            yield km
        finally:
            await km.shutdown_all(now=True)


# Test that it works with a max of 4
class TestLimitedKernelManagerApplied(TestAsyncKernelManager):
    __test__ = True

    # static so picklable for multiprocessing on Windows
    @staticmethod
    @asynccontextmanager
    async def _get_tcp_km():
        c = Config()
        c.LimitedKernelManager.max_kernels = 4
        km = LimitedKernelManager(config=c)
        try:
            yield km
        finally:
            await km.shutdown_all()

    @gen_test(timeout=60)
    async def test_touch_max(self):
        async with self._get_tcp_km() as km:
            kids = []
            for i in range(4):
                kid = await km.start_kernel(stdout=PIPE, stderr=PIPE)
                self.assertIn(kid, km)
                kids.append(kid)

            # Remove and add one to make sure we don't count closed kernels
            await km.shutdown_kernel(kid)
            self.assertNotIn(kid, km)
            kids.pop()

            kid = await km.start_kernel(stdout=PIPE, stderr=PIPE)
            self.assertIn(kid, km)
            kids.append(kid)

            await km.shutdown_all()
            for kid in kids:
                self.assertNotIn(kid, km)

    @gen_test(timeout=60)
    async def test_breach_max(self):
        async with self._get_tcp_km() as km:
            kids = []
            for i in range(4):
                kid = await km.start_kernel(stdout=PIPE, stderr=PIPE)
                self.assertIn(kid, km)
                kids.append(kid)

            with self.assertRaises(MaximumKernelsException):
                await km.start_kernel(stdout=PIPE, stderr=PIPE)

            # Remove and add one to make sure we correctly recovered
            await km.shutdown_kernel(kid)
            self.assertNotIn(kid, km)
            kids.pop()

            kid = await km.start_kernel(stdout=PIPE, stderr=PIPE)
            self.assertIn(kid, km)
            kids.append(kid)

            await km.shutdown_all()
            for kid in kids:
                self.assertNotIn(kid, km)



python_update_cwd_code = """
import os;
os.chdir({cwd!r});
del os;
"""

python_update_env_code = """
import os;
for key, value in {env!r}.items():
    os.environ[key] = value;
del key
del value
del os;
"""

python_init_import_code = """
import importlib
for name in {modules!r}:
    importlib.import_module(name)
    del name
del importlib
"""



import asyncio
from contextlib import asynccontextmanager
from subprocess import PIPE
from unittest import TestCase

from jupyter_client.kernelspec import NATIVE_KERNEL_NAME
import pytest
from tornado.testing import AsyncTestCase, gen_test
from traitlets.config.loader import Config

try:
    from .. import (
        PooledKernelManager,
        MaximumKernelsException,
    )
except ImportError:
    pass

from .utils import async_shutdown_all_direct, TestAsyncKernelManager

# Test that it works as normal with default config
class TestPooledKernelManagerUnused(TestAsyncKernelManager):
    __test__ = True

    # static so picklable for multiprocessing on Windows
    @staticmethod
    @asynccontextmanager
    async def _get_tcp_km():
        c = Config()
        km = PooledKernelManager(config=c)
        try:
            yield km
        finally:
            await km.shutdown_all(now=True)


# Test that it works with an unstrict pool
class TestPooledKernelManagerApplied(TestAsyncKernelManager):
    __test__ = True

    # static so picklable for multiprocessing on Windows
    @staticmethod
    @asynccontextmanager
    async def _get_tcp_km():
        c = Config()
        c.LimitedKernelManager.max_kernels = 4
        c.PooledKernelManager.fill_delay = 0
        c.PooledKernelManager.kernel_pools = {NATIVE_KERNEL_NAME: 2}
        c.PooledKernelManager.pool_kwargs = {NATIVE_KERNEL_NAME: dict(stdout=PIPE, stderr=PIPE)}
        km = PooledKernelManager(config=c)
        try:
            await km.wait_for_pool()
            yield km
        finally:
            await km.shutdown_all()

    @gen_test(timeout=60)
    async def test_exceed_pool_size(self):
        async with self._get_tcp_km() as km:
            self.assertEqual(len(km._pools[NATIVE_KERNEL_NAME]), 2)
            kids = []
            for i in range(4):
                kid = await km.start_kernel(stdout=PIPE, stderr=PIPE)
                self.assertIn(kid, km)
                kids.append(kid)

            await async_shutdown_all_direct(km)
            for kid in kids:
                self.assertNotIn(kid, km)

            # Cycle again to assure the pool survives that
            kids = []
            for i in range(4):
                kid = await km.start_kernel(stdout=PIPE, stderr=PIPE)
                self.assertIn(kid, km)
                kids.append(kid)

            await km.shutdown_all()
            for kid in kids:
                self.assertNotIn(kid, km)

    @gen_test
    async def test_decrease_pool_size(self):
        async with self._get_tcp_km() as km:
            km.kernel_pools = {NATIVE_KERNEL_NAME: 1}
            self.assertEqual(len(km._pools[NATIVE_KERNEL_NAME]), 1)
            # km.shutdown_kernel is not reentrant, so await:
            await asyncio.gather(*km._discarded)

    @gen_test
    async def test_increase_pool_size(self):
        async with self._get_tcp_km() as km:
            km.kernel_pools = {NATIVE_KERNEL_NAME: 3}
            self.assertEqual(len(km._pools[NATIVE_KERNEL_NAME]), 3)

    @gen_test(timeout=60)
    async def test_breach_max(self):
        async with self._get_tcp_km() as km:
            kids = []
            for i in range(4):
                kid = await km.start_kernel(stdout=PIPE, stderr=PIPE)
                self.assertIn(kid, km)
                kids.append(kid)

            with self.assertRaises(MaximumKernelsException):
                await km.start_kernel(stdout=PIPE, stderr=PIPE)

            # Remove and add one to make sure we correctly recovered
            await km.shutdown_kernel(kid)
            self.assertNotIn(kid, km)
            kids.pop()

            kid = await km.start_kernel(stdout=PIPE, stderr=PIPE)
            self.assertIn(kid, km)
            kids.append(kid)

            await km.shutdown_all()
            for kid in kids:
                self.assertNotIn(kid, km)
            # shutdown again is okay, because we have no kernels
            await km.shutdown_all()


# Test that it works with an strict pool
class TestPooledKernelManagerStrict(AsyncTestCase):
    @gen_test
    async def test_strict_name_correct(self):
        c = Config()
        c.PooledKernelManager.kernel_pools = {NATIVE_KERNEL_NAME: 1}
        c.PooledKernelManager.strict_pool_names = True
        km = PooledKernelManager(config=c)

        try:
            kid = await km.start_kernel(kernel_name=NATIVE_KERNEL_NAME, stdout=PIPE, stderr=PIPE)
            self.assertIn(kid, km)

        finally:
            await km.shutdown_all()
        self.assertNotIn(kid, km)

    @gen_test
    async def test_strict_name_incorrect(self):
        c = Config()
        c.PooledKernelManager.kernel_pools = {NATIVE_KERNEL_NAME: 1}
        c.PooledKernelManager.strict_pool_names = True
        km = PooledKernelManager(config=c)

        try:
            with self.assertRaisesRegex(ValueError, "Cannot start kernel with name"):
                kid = await km.start_kernel(kernel_name="foo", stdout=PIPE, stderr=PIPE)
            self.assertEqual(len(km._pools[NATIVE_KERNEL_NAME]), 1)
        finally:
            await km.shutdown_all()

    @gen_test
    async def test_strict_kwargs_correct(self):
        c = Config()
        c.PooledKernelManager.kernel_pools = {NATIVE_KERNEL_NAME: 1}
        c.PooledKernelManager.pool_kwargs = {NATIVE_KERNEL_NAME: dict(stdout=PIPE, stderr=PIPE)}
        c.PooledKernelManager.strict_pool_kwargs = True
        km = PooledKernelManager(config=c)

        try:
            kid = await km.start_kernel(stdout=PIPE, stderr=PIPE)
            self.assertIn(kid, km)
        finally:
            await km.shutdown_all()
        self.assertNotIn(kid, km)

    @gen_test
    async def test_strict_kwargs_incorrect(self):
        c = Config()
        c.PooledKernelManager.kernel_pools = {NATIVE_KERNEL_NAME: 1}
        c.PooledKernelManager.pool_kwargs = {NATIVE_KERNEL_NAME: dict(stdout=PIPE, stderr=PIPE)}
        c.PooledKernelManager.strict_pool_kwargs = True
        km = PooledKernelManager(config=c)

        try:
            with self.assertRaisesRegex(ValueError, "Cannot start kernel with kwargs"):
                kid = await km.start_kernel()
            self.assertEqual(len(km._pools[NATIVE_KERNEL_NAME]), 1)
        finally:
            await km.shutdown_all()

    @gen_test
    async def test_both_strict_correct(self):
        c = Config()
        c.PooledKernelManager.kernel_pools = {NATIVE_KERNEL_NAME: 1}
        c.PooledKernelManager.strict_pool_names = True
        c.PooledKernelManager.pool_kwargs = {NATIVE_KERNEL_NAME: dict(stdout=PIPE, stderr=PIPE)}
        c.PooledKernelManager.strict_pool_kwargs = True
        km = PooledKernelManager(config=c)

        try:
            kid = await km.start_kernel(kernel_name=NATIVE_KERNEL_NAME, stdout=PIPE, stderr=PIPE)
            self.assertIn(kid, km)
        finally:
            await km.shutdown_all()
        self.assertNotIn(kid, km)



import asyncio
import os
import sys

import pytest

# Mostly for CI, but useful for everyone
os.environ["ASYNC_TEST_TIMEOUT"] = "30"

if os.name == "nt" and sys.version_info >= (3, 7):
    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())


@pytest.fixture
def event_loop():
    # Make sure we test against a selector event loop
    # since pyzmq doesn't like the proactor loop.
    # This fixture is picked up by pytest-asyncio
    if os.name == "nt" and sys.version_info >= (3, 7):
        asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
    loop = asyncio.SelectorEventLoop()
    try:
        yield loop
    finally:
        loop.close()



# coding: utf-8

# Copyright (c) Vidar Tonaas Fauske.
# Distributed under the terms of the Modified BSD License.
"""Hotpot - Jupyter kernel manager helpers
"""

from ._version import __version__

from .limited import MaximumKernelsException, SyncLimitedKernelManager
from .pooled_sync import SyncPooledKernelManager

__all__ = [
    "__version__",
    "MaximumKernelsException",
    "SyncPooledKernelManager",
    "SyncLimitedKernelManager",
]

try:
    from .pooled import PooledKernelManager

    __all__.append("PooledKernelManager")
except ImportError:
    pass

try:
    from .limited import LimitedKernelManager

    __all__.append("LimitedKernelManager")
except ImportError:
    pass

try:
    from .mapping import PooledMappingKernelManager

    __all__.append("PooledMappingKernelManager")
except ImportError:
    pass



from jupyter_server.services.kernels.kernelmanager import MappingKernelManager

from .pooled_sync import SyncPooledKernelManager


class SyncPooledMappingKernelManager(SyncPooledKernelManager, MappingKernelManager):
    def restart_kernel(self, kernel_id, **kwargs):
        if kwargs:
            self.log.warning("Ignored arguments to restart_kernel: %r", kwargs)
        return super().restart_kernel(kernel_id)



from jupyter_server.services.kernels.kernelmanager import AsyncMappingKernelManager

from .pooled import PooledKernelManager


class PooledMappingKernelManager(PooledKernelManager, AsyncMappingKernelManager):
    async def restart_kernel(self, kernel_id, **kwargs):
        if kwargs:
            self.log.warning("Ignored arguments to restart_kernel: %r", kwargs)
        return await super().restart_kernel(kernel_id)



# coding: utf-8

# Copyright (c) Vidar Tonaas Fauske.
# Distributed under the terms of the Modified BSD License.
"""Hotpot - Jupyter kernel manager helpers

This module contains
"""

from jupyter_client.multikernelmanager import MultiKernelManager

from traitlets import Integer


class MaximumKernelsException(Exception):
    pass


class SyncLimitedKernelManager(MultiKernelManager):
    max_kernels = Integer(
        0,
        config=True,
        help="The maximum number of concurrent kernels",
    )

    def start_kernel(self, kernel_name=None, **kwargs):
        if "kernel_id" not in kwargs and len(self) >= self.max_kernels > 0:
            self.log.debug("Refusing to start kernel, maximum number reached.")
            raise MaximumKernelsException("No kernels are available.")
        return super().start_kernel(kernel_name=kernel_name, **kwargs)


__all__ = [
    "MaximumKernelsException",
    "SyncLimitedKernelManager",
]


try:
    from jupyter_client.multikernelmanager import (
        AsyncMultiKernelManager,
        MultiKernelManager,
    )

    class LimitedKernelManager(AsyncMultiKernelManager):
        max_kernels = Integer(
            0,
            config=True,
            help="The maximum number of concurrent kernels",
        )

        def pre_start_kernel(self, kernel_name, kwargs):
            if len(self) >= self.max_kernels > 0:
                self.log.debug("Refusing to start kernel, maximum number reached.")
                raise MaximumKernelsException("No kernels are available.")
            return super().pre_start_kernel(kernel_name, kwargs)

    __all__.append("LimitedKernelManager")

except ImportError:
    pass



from contextlib import asynccontextmanager
from subprocess import PIPE

from jupyter_client.kernelspec import NATIVE_KERNEL_NAME
from pytest import mark
from traitlets.config.loader import Config
from tornado.testing import gen_test

from .. import MaximumKernelsException

try:
    from ..mapping_sync import SyncPooledMappingKernelManager
except ImportError as e:
    print(f"Won't be able to test synced pool: {e}")

from ..async_utils import ensure_async
from .utils import async_shutdown_all_direct, TestAsyncKernelManager


# Test that it works as normal with default config
class TestSyncMappingKernelManagerUnused(TestAsyncKernelManager):
    __test__ = True

    # static so picklable for multiprocessing on Windows
    @staticmethod
    @asynccontextmanager
    async def _get_tcp_km():
        c = Config()
        km = SyncPooledMappingKernelManager(config=c)
        try:
            yield km
        finally:
            await ensure_async(km.shutdown_all())

    # Mapping manager doesn't handle this:
    @mark.skip()
    @gen_test
    async def test_tcp_lifecycle_with_kernel_id(self):
        pass


# Test that it works with a max that is larger than pool size
class TestSyncMappingKernelManagerApplied(TestAsyncKernelManager):
    __test__ = True

    # static so picklable for multiprocessing on Windows
    @staticmethod
    @asynccontextmanager
    async def _get_tcp_km():
        c = Config()
        c.SyncLimitedKernelManager.max_kernels = 4
        c.SyncPooledMappingKernelManager.fill_delay = 0
        c.SyncPooledMappingKernelManager.kernel_pools = {NATIVE_KERNEL_NAME: 2}
        c.SyncPooledMappingKernelManager.pool_kwargs = {
            NATIVE_KERNEL_NAME: dict(stdout=PIPE, stderr=PIPE)
        }
        km = SyncPooledMappingKernelManager(config=c)
        try:
            yield km
        finally:
            await ensure_async(km.shutdown_all())

    # Mapping manager doesn't handle this:
    @mark.skip()
    @gen_test
    async def test_tcp_lifecycle_with_kernel_id(self):
        pass

    @gen_test(timeout=60)
    async def test_exceed_pool_size(self):
        async with self._get_tcp_km() as km:
            self.assertEqual(len(km._pools[NATIVE_KERNEL_NAME]), 2)
            kids = []
            for i in range(4):
                kid = await ensure_async(km.start_kernel(stdout=PIPE, stderr=PIPE))
                self.assertIn(kid, km)
                kids.append(kid)

            await async_shutdown_all_direct(km)
            for kid in kids:
                self.assertNotIn(kid, km)

            # Cycle again to assure the pool survives that
            kids = []
            for i in range(4):
                kid = await ensure_async(km.start_kernel(stdout=PIPE, stderr=PIPE))
                self.assertIn(kid, km)
                kids.append(kid)

            await ensure_async(km.shutdown_all())
            for kid in kids:
                self.assertNotIn(kid, km)

    @gen_test(timeout=60)
    async def test_breach_max(self):
        async with self._get_tcp_km() as km:
            kids = []
            for i in range(4):
                kid = await ensure_async(km.start_kernel(stdout=PIPE, stderr=PIPE))
                self.assertIn(kid, km)
                kids.append(kid)

            with self.assertRaises(MaximumKernelsException):
                await ensure_async(km.start_kernel(stdout=PIPE, stderr=PIPE))

            # Remove and add one to make sure we correctly recovered
            await ensure_async(km.shutdown_kernel(kid))
            self.assertNotIn(kid, km)
            kids.pop()

            kid = await ensure_async(km.start_kernel(stdout=PIPE, stderr=PIPE))
            self.assertIn(kid, km)
            kids.append(kid)

            await ensure_async(km.shutdown_all())
            for kid in kids:
                self.assertNotIn(kid, km)



import asyncio
from contextlib import asynccontextmanager
from unittest import mock
from pathlib import Path
from subprocess import PIPE
from tempfile import TemporaryDirectory
from unittest import TestCase

from jupyter_client.kernelspec import NATIVE_KERNEL_NAME
import pytest
from tornado.testing import AsyncTestCase, gen_test
from traitlets.config.loader import Config

from ..client_helper import ExecClient

try:
    from .. import (
        PooledKernelManager,
        PooledMappingKernelManager,
        MaximumKernelsException,
    )
except ImportError:
    pass

from .utils import async_shutdown_all_direct


class TestInitializePooled(AsyncTestCase):

    # This is a snaity check for test_imports
    @gen_test(timeout=20)
    async def test_imports_not_used(self):
        c = Config()
        c.PooledKernelManager.kernel_pools = {NATIVE_KERNEL_NAME: 1}
        c.PooledKernelManager.strict_pool_names = True
        c.python_imports = []
        km = PooledKernelManager(config=c)

        try:
            kid = await km.start_kernel()
            kernel = km.get_kernel(kid)
            self.assertIsNotNone(kernel)

            client = ExecClient(kernel, _store_outputs=True)
            async with client.setup_kernel():
                await client.execute('import sys\nprint("turtle" in sys.modules)')
            self.assertEqual(
                client._outputs,
                [{"name": "stdout", "output_type": "stream", "text": "False\n"}],
            )
        finally:
            await km.shutdown_all()

    @gen_test(timeout=20)
    async def test_imports(self):
        c = Config()
        c.PooledKernelManager.kernel_pools = {NATIVE_KERNEL_NAME: 1}
        c.PooledKernelManager.strict_pool_names = True
        c.PooledKernelManager.python_imports = ["turtle"]
        # Need mapping to resolve kernel language
        km = PooledKernelManager(config=c)

        try:
            kid = await km.start_kernel()
            kernel = km.get_kernel(kid)
            self.assertIsNotNone(kernel)

            client = ExecClient(kernel, _store_outputs=True)
            async with client.setup_kernel():
                await client.execute('import sys\nprint("turtle" in sys.modules)')
            self.assertEqual(
                client._outputs,
                [{"name": "stdout", "output_type": "stream", "text": "True\n"}],
            )
        finally:
            await km.shutdown_all()

    @pytest.mark.xfail()  # initialize happens before update, so this won't work
    @gen_test(timeout=20)
    async def test_cwd_import(self):
        with TemporaryDirectory() as tmp_dir:
            # create a lcoal py module in cwd to test import
            foo_mod = Path(tmp_dir) / "foo_module.py"
            foo_mod.write_text("foo = 1")

            c = Config()
            c.PooledKernelManager.kernel_pools = {NATIVE_KERNEL_NAME: 1}
            c.PooledKernelManager.strict_pool_names = True
            c.PooledKernelManager.python_imports = ["foo_module"]
            km = PooledKernelManager(config=c)

            try:
                kid = await km.start_kernel(cwd=tmp_dir)
                kernel = km.get_kernel(kid)
                self.assertIsNotNone(kernel)

                client = ExecClient(kernel, _store_outputs=True)
                async with client.setup_kernel():
                    await client.execute('import sys\nprint("foo_module" in sys.modules)')
                self.assertEqual(
                    client._outputs,
                    [{"name": "stdout", "output_type": "stream", "text": "True\n"}],
                )
            finally:
                await km.shutdown_all()

    @gen_test(timeout=20)
    async def test_file_init(self):
        with TemporaryDirectory() as tmp_dir:
            c = Config()
            c.PooledKernelManager.kernel_pools = {NATIVE_KERNEL_NAME: 1}
            c.PooledKernelManager.strict_pool_names = True
            # Need mapping manager to resolve init files:
            km = PooledKernelManager(config=c)

            with mock.patch("jupyter_core.paths.jupyter_config_path") as jcp:
                jcp.return_value = [tmp_dir]
                # create a local py module in cwd to test import
                foo_mod = Path(tmp_dir) / f"kernel_pool_init_{NATIVE_KERNEL_NAME}.py"
                foo_mod.write_text("foo = 1")

                try:
                    kid = await km.start_kernel()
                    kernel = km.get_kernel(kid)
                    self.assertIsNotNone(kernel)

                    client = ExecClient(kernel, _store_outputs=True)
                    async with client.setup_kernel():
                        await client.execute("print(foo)")
                    self.assertEqual(
                        client._outputs,
                        [{"name": "stdout", "output_type": "stream", "text": "1\n"}],
                    )
                finally:
                    await km.shutdown_all()

    @gen_test(timeout=20)
    async def test_init(self):
        with TemporaryDirectory() as tmp_dir:
            c = Config()
            c.PooledKernelManager.kernel_pools = {NATIVE_KERNEL_NAME: 1}
            c.PooledKernelManager.strict_pool_names = True
            c.PooledKernelManager.initialization_code = {NATIVE_KERNEL_NAME: "import turtle"}
            # Need mapping manager to resolve init files:
            km = PooledKernelManager(config=c)
            await km.wait_for_pool()

        try:
            kid = await km.start_kernel()
            kernel = km.get_kernel(kid)
            self.assertIsNotNone(kernel)

            client = ExecClient(kernel, _store_outputs=True)
            async with client.setup_kernel():
                await client.execute('import sys\nprint("turtle" in sys.modules)')
            self.assertEqual(
                client._outputs,
                [{"name": "stdout", "output_type": "stream", "text": "True\n"}],
            )
        finally:
            await km.shutdown_all()


import asyncio
import threading
import uuid
import multiprocessing as mp

import pytest
from subprocess import PIPE
from unittest import TestCase

from jupyter_client import KernelManager
from jupyter_client.tests.utils import skip_win32
from jupyter_client.localinterfaces import localhost


def shutdown_all_direct(km):
    kids = km.list_kernel_ids()
    futs = []
    for kid in kids:
        km.shutdown_kernel(kid)


class TestKernelManager(TestCase):
    # Prevent the base class from being collected directly
    __test__ = False

    # static so picklable for multiprocessing on Windows
    @staticmethod
    def _run_lifecycle(km, test_kid=None):
        if test_kid:
            kid = km.start_kernel(stdout=PIPE, stderr=PIPE, kernel_id=test_kid)
            assert kid == test_kid
        else:
            kid = km.start_kernel(stdout=PIPE, stderr=PIPE)
        assert km.is_alive(kid)
        assert kid in km
        assert kid in km.list_kernel_ids()
        km.restart_kernel(kid, now=True)
        assert km.is_alive(kid)
        assert kid in km.list_kernel_ids()
        km.interrupt_kernel(kid)
        k = km.get_kernel(kid)
        assert isinstance(k, KernelManager)
        km.shutdown_kernel(kid, now=True)
        assert kid not in km, f"{kid} not in {km}"

    def _run_cinfo(self, km, transport, ip):
        kid = km.start_kernel(stdout=PIPE, stderr=PIPE)
        k = km.get_kernel(kid)
        cinfo = km.get_connection_info(kid)
        self.assertEqual(transport, cinfo["transport"])
        self.assertEqual(ip, cinfo["ip"])
        self.assertTrue("stdin_port" in cinfo)
        self.assertTrue("iopub_port" in cinfo)
        stream = km.connect_iopub(kid)
        stream.close()
        self.assertTrue("shell_port" in cinfo)
        stream = km.connect_shell(kid)
        stream.close()
        self.assertTrue("hb_port" in cinfo)
        stream = km.connect_hb(kid)
        stream.close()
        km.shutdown_kernel(kid, now=True)
        self.assertNotIn(kid, km)

    def test_tcp_lifecycle(self):
        self.raw_tcp_lifecycle()

    def test_tcp_lifecycle_with_kernel_id(self):
        self.raw_tcp_lifecycle(test_kid=str(uuid.uuid4()))

    def test_shutdown_all(self):
        with self._get_tcp_km() as km:
            kid = km.start_kernel(stdout=PIPE, stderr=PIPE)
            self.assertIn(kid, km)
            km.shutdown_all()
            self.assertNotIn(kid, km)
            # shutdown again is okay, because we have no kernels
            km.shutdown_all()

    def test_tcp_cinfo(self):
        with self._get_tcp_km() as km:
            self._run_cinfo(km, "tcp", localhost())

    def test_start_sequence_tcp_kernels(self):
        """Ensure that a sequence of kernel startups doesn't break anything."""
        with self._get_tcp_km() as km:
            self._run_lifecycle(km)
        with self._get_tcp_km() as km:
            self._run_lifecycle(km)
        with self._get_tcp_km() as km:
            self._run_lifecycle(km)

    # static so picklable for multiprocessing on Windows
    @classmethod
    def tcp_lifecycle_with_loop(cls):
        # Ensure each thread has an event loop
        import os, sys

        if os.name == "nt" and sys.version_info >= (3, 7):
            asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
        asyncio.set_event_loop(asyncio.new_event_loop())
        cls.raw_tcp_lifecycle()

    # static so picklable for multiprocessing on Windows
    @classmethod
    def raw_tcp_lifecycle(cls, test_kid=None):
        with cls._get_tcp_km() as km:
            cls._run_lifecycle(km, test_kid=test_kid)

    def test_start_parallel_thread_kernels(self):
        self.raw_tcp_lifecycle()

        thread = threading.Thread(target=self.tcp_lifecycle_with_loop)
        thread2 = threading.Thread(target=self.tcp_lifecycle_with_loop)
        try:
            thread.start()
            thread2.start()
        finally:
            thread.join()
            thread2.join()

    @pytest.mark.skipif(mp.get_start_method() == "fork", reason="Fork")
    def test_start_parallel_process_kernels(self):
        self.raw_tcp_lifecycle()

        thread = threading.Thread(target=self.tcp_lifecycle_with_loop)
        proc = mp.Process(target=self.tcp_lifecycle_with_loop)

        try:
            thread.start()
            proc.start()
        finally:
            proc.join()
            thread.join()

        assert proc.exitcode == 0



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

# This file is a paired down version of nbclient, so that
# it just a wrapper around jupyter_client that responds correctly to
# certain events.

import atexit
import collections
import datetime
import base64
import signal
from textwrap import dedent

from async_generator import asynccontextmanager
from contextlib import contextmanager

from time import monotonic
from queue import Empty
import asyncio
import typing as t

from traitlets.config.configurable import LoggingConfigurable
from traitlets import List, Unicode, Bool, Enum, Any, Type, Dict, Integer, default

from nbformat.v4 import output_from_msg
from jupyter_client import KernelManager
from jupyter_client.client import KernelClient

from .async_utils import run_sync, ensure_async


class ControlSignal(Exception):
    pass


class ExecTimeoutError(TimeoutError, ControlSignal):
    pass


class ExecutionError(ControlSignal):
    def __init__(self, traceback: str, ename: str, evalue: str) -> None:
        super().__init__(traceback)
        self.traceback = traceback
        self.ename = ename
        self.evalue = evalue

    def __reduce__(self) -> tuple:
        return type(self), (self.traceback, self.ename, self.evalue)

    def __str__(self) -> str:
        s = self.__unicode__()
        if not isinstance(s, str):
            s = s.encode("utf8", "replace")
        return s

    def __unicode__(self) -> str:
        return self.traceback

    @classmethod
    def from_msg(cls, msg: Dict):
        """Instantiate from message contents (message is either execute_reply or error)"""
        tb = "\n".join(msg.get("traceback", []))
        return cls(
            traceback=tb,
            ename=msg.get("ename", "<Error>"),
            evalue=msg.get("evalue", ""),
        )


class DeadKernelError(RuntimeError):
    pass


def timestamp() -> str:
    return datetime.datetime.utcnow().isoformat() + "Z"


class ExecClient(LoggingConfigurable):
    """
    A client for executing code on a Jupyter kernel
    """

    timeout: int = Integer(
        None,
        allow_none=True,
        help=dedent(
            """
            The time to wait (in seconds) for output from executions.
            If an execution takes longer, a TimeoutError is raised.

            ``None`` or ``-1`` will disable the timeout. If ``timeout_func`` is set,
            it overrides ``timeout``.
            """
        ),
    ).tag(config=True)

    startup_timeout: int = Integer(
        60,
        help=dedent(
            """
            The time to wait (in seconds) for the kernel to start.
            If kernel startup takes longer, a RuntimeError is
            raised.
            """
        ),
    ).tag(config=True)

    iopub_timeout: int = Integer(
        4,
        allow_none=False,
        help=dedent(
            """
            The time to wait (in seconds) for IOPub output. This generally
            doesn't need to be set, but on some slow networks (such as CI
            systems) the default timeout might not be long enough to get all
            messages.
            """
        ),
    ).tag(config=True)

    shell_timeout_interval: int = Integer(
        5,
        allow_none=False,
        help=dedent(
            """
            The time to wait (in seconds) for Shell output before retrying.
            This generally doesn't need to be set, but if one needs to check
            for dead kernels at a faster rate this can help.
            """
        ),
    ).tag(config=True)

    def __init__(
        self, km: KernelManager = None, _store_outputs: bool = False, **kw  # for testing purposes
    ) -> None:
        """Initializes the execution manager.

        Parameters
        ----------
        km : KernelManager (optional)
            Optional kernel manager. If none is provided, a kernel manager will
            be created.
        """
        super().__init__(**kw)
        self.km: KernelManager = km
        self.kc: t.Optional[KernelClient] = None
        self._store_outputs = _store_outputs
        self._outputs = []

    async def cleanup_client(self) -> None:
        if getattr(self, "kc") and self.kc is not None:
            await ensure_async(self.kc.stop_channels())
            self.kc = None

    async def _cleanup_kernel(self) -> None:
        assert self.km is not None
        try:
            # Queue the manager to kill the process, and recover gracefully if it's already dead.
            if await ensure_async(self.km.is_alive()):
                await ensure_async(self.km.shutdown_kernel())
        except RuntimeError as e:
            # The error isn't specialized, so we have to check the message
            if "No kernel is running!" not in str(e):
                raise
        finally:
            # Remove any state left over even if we failed to stop the kernel
            await self.cleanup_client()
            await ensure_async(self.km.cleanup())
            self.km = None

    _sync_cleanup_kernel = run_sync(_cleanup_kernel)

    async def start_new_kernel_client(self) -> KernelClient:
        """Creates a new kernel client.

        Returns
        -------
        kc : KernelClient
            Kernel client as created by the kernel manager ``km``.
        """
        assert self.km is not None
        self.kc = self.km.client()
        await ensure_async(self.kc.start_channels())
        try:
            await ensure_async(self.kc.wait_for_ready(timeout=self.startup_timeout))
        except RuntimeError:
            await self._cleanup_kernel()
            raise
        self.kc.allow_stdin = False
        return self.kc

    async def ensure_kernel_client(self) -> None:
        """Ensure there is a ready kernel client awailable for use."""
        if self.kc is None:
            self.kc = await self.start_new_kernel_client()
        else:
            try:
                await ensure_async(self.kc.wait_for_ready(timeout=self.startup_timeout))
            except RuntimeError:
                await self._cleanup_kernel()
                raise

    @asynccontextmanager
    async def setup_kernel(self) -> t.AsyncGenerator:
        """
        Context manager for setting up the kernel to execute a notebook.

        This assigns the Kernel Client(``self.kc``) if missing .

        When control returns from the yield it stops the client's zmq channels.

        Handlers for SIGINT and SIGTERM are also added to cleanup in case of unexpected shutdown.
        """

        # self._sync_cleanup_kernel uses run_async, which ensures the ioloop is running again.
        # This is necessary as the ioloop has stopped once atexit fires.
        atexit.register(self._sync_cleanup_kernel)

        def on_signal():
            asyncio.ensure_future(self._cleanup_kernel())
            atexit.unregister(self._sync_cleanup_kernel)

        loop = asyncio.get_event_loop()
        try:
            loop.add_signal_handler(signal.SIGINT, on_signal)
            loop.add_signal_handler(signal.SIGTERM, on_signal)
        except (NotImplementedError, RuntimeError):
            # NotImplementedError: Windows does not support signals.
            # RuntimeError: Raised when add_signal_handler is called outside the main thread
            pass

        try:
            await self.ensure_kernel_client()
            yield
        finally:
            atexit.unregister(self._sync_cleanup_kernel)
            try:
                loop.remove_signal_handler(signal.SIGINT)
                loop.remove_signal_handler(signal.SIGTERM)
            except (NotImplementedError, RuntimeError):
                pass
            await self.cleanup_client()

    async def execute(self, source: str, **kwargs) -> t.Optional[dict]:
        """
        Executes code. Requires that a setup_kernel context is held.

        Parameters
        ----------
        source :
            The source code to execute.
        kwargs :
            Any option for ``self.kernel_manager_class.start_kernel()``. Because
            that defaults to AsyncKernelManager, this will likely include options
            accepted by ``jupyter_client.AsyncKernelManager.start_kernel()``,
            which includes ``cwd``.

            ``reset_kc`` if True, the kernel client will be reset and a new one
            will be created (default: False).

        Returns
        -------
        The execute reply message, or None if skipped bc of empty code.

        Raises
        ------
        CellExecutionError
            If execution failed and should raise an exception, this will be raised
            with defaults about the failure.
        """

        assert self.kc is not None
        if not source.strip():
            self.log.debug("Skipping empty code")
            return None

        self.log.debug("Executing code")

        parent_msg_id = await ensure_async(
            self.kc.execute(
                source,
                store_history=False,
            )
        )
        exec_timeout = self._get_timeout()

        task_poll_kernel_alive = asyncio.ensure_future(self._poll_kernel_alive())
        task_poll_output_msg = asyncio.ensure_future(self._poll_output_msg(parent_msg_id))
        self.task_poll_for_reply = asyncio.ensure_future(
            self._poll_for_reply(
                parent_msg_id,
                exec_timeout,
                task_poll_output_msg,
                task_poll_kernel_alive,
            )
        )
        try:
            exec_reply = await self.task_poll_for_reply
        except asyncio.CancelledError:
            # can only be cancelled by task_poll_kernel_alive when the kernel is dead
            task_poll_output_msg.cancel()
            raise DeadKernelError("Kernel died")
        except Exception as e:
            # Best effort to cancel request if it hasn't been resolved
            try:
                # Check if the task_poll_output is doing the raising for us
                if not isinstance(e, ControlSignal):
                    task_poll_output_msg.cancel()
            finally:
                raise

        self._check_raise_for_error(exec_reply)
        return exec_reply

    async def _poll_for_reply(
        self,
        msg_id: str,
        timeout: t.Optional[int],
        task_poll_output_msg: asyncio.Future,
        task_poll_kernel_alive: asyncio.Future,
    ) -> t.Dict:

        assert self.kc is not None
        new_timeout: t.Optional[float] = None
        if timeout is not None:
            deadline = monotonic() + timeout
            new_timeout = float(timeout)
        while True:
            try:
                msg = await ensure_async(self.kc.shell_channel.get_msg(timeout=new_timeout))
                if msg["parent_header"].get("msg_id") == msg_id:
                    try:
                        await asyncio.wait_for(task_poll_output_msg, self.iopub_timeout)
                    except (asyncio.TimeoutError, Empty):
                        if self.raise_on_iopub_timeout:
                            task_poll_kernel_alive.cancel()
                            raise ExecTimeoutError("Timeout waiting for IOPub output")
                        else:
                            self.log.warning("Timeout waiting for IOPub output")
                    task_poll_kernel_alive.cancel()
                    return msg
                else:
                    if new_timeout is not None:
                        new_timeout = max(0, deadline - monotonic())
            except Empty:
                # received no message, check if kernel is still alive
                assert timeout is not None
                task_poll_kernel_alive.cancel()
                await self._check_alive()
                await self._handle_timeout(timeout)

    async def _poll_output_msg(self, parent_msg_id: str) -> None:

        assert self.kc is not None
        while True:
            msg = await ensure_async(self.kc.iopub_channel.get_msg(timeout=None))
            if msg["parent_header"].get("msg_id") == parent_msg_id:
                if self.process_message(msg):
                    return

    async def _poll_kernel_alive(self) -> None:
        while True:
            await asyncio.sleep(1)
            try:
                await self._check_alive()
            except DeadKernelError:
                assert self.task_poll_for_reply is not None
                self.task_poll_for_reply.cancel()
                return

    def _get_timeout(self) -> int:
        timeout = self.timeout

        if not timeout or timeout < 0:
            timeout = None

        return timeout

    async def _handle_timeout(self, timeout: int) -> None:

        self.log.error("Timeout waiting for execute reply (%is)." % timeout)
        if self.interrupt_on_timeout:
            self.log.error("Interrupting kernel")
            assert self.km is not None
            await ensure_async(self.km.interrupt_kernel())
        else:
            raise ExecTimeoutError("Execution timed out")

    async def _check_alive(self) -> None:
        assert self.kc is not None
        if not await ensure_async(self.kc.is_alive()):
            self.log.error("Kernel died while waiting for execute reply.")
            raise DeadKernelError("Kernel died")

    def _check_raise_for_error(self, exec_reply: t.Optional[t.Dict]) -> None:

        if exec_reply is None:
            return None

        exec_reply_content = exec_reply["content"]
        if exec_reply_content["status"] != "error":
            return None

        raise ExecutionError.from_msg(exec_reply_content)

    def process_message(self, msg: t.Dict) -> bool:
        """
        Processes a kernel message, updates cell state, and returns the
        resulting output object that was appended to cell.outputs.

        The input argument *cell* is modified in-place.

        Parameters
        ----------
        msg : dict
            The kernel message being processed.

        Returns
        -------
        Whether the message indicates computation completeness.

        """
        msg_type = msg["msg_type"]
        self.log.debug("msg_type: %s", msg_type)
        content = msg["content"]
        if msg_type == "status":
            if content["execution_state"] == "idle":
                return True
        elif self._store_outputs and msg_type not in [
            "clear_output",
            "comm",
            "execute_input",
            "update_display_data",
        ]:
            # Assign output as our processed "result"
            self.output(self._outputs, msg)
        return False

    def output(self, outs: t.List, msg: t.Dict) -> t.Optional[t.List]:

        try:
            out = output_from_msg(msg)
        except ValueError:
            self.log.error("unhandled iopub msg: " + msg["msg_type"])
            return

        outs.append(out)



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['hotpot_km', 'hotpot_km.tests']

package_data = \
{'': ['*']}

install_requires = \
['async_generator', 'jupyter_client', 'nbformat', 'nest_asyncio', 'traitlets']

extras_require = \
{'mapping': ['jupyter_server'],
 'test': ['pytest', 'pytest-cov', 'jupyter_server', 'jupyter_client[test]']}

setup(name='hotpot_km',
      version='0.2.1',
      description='Hotpot - Jupyter kernel manager helpers',
      author='Vidar Tonaas Fauske',
      author_email='vidartf@gmail.com',
      url='https://github.com/voila-dashboards/hotpot_km',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      python_requires='~=3.7',
     )



import asyncio
from contextlib import asynccontextmanager
from pathlib import Path
from subprocess import PIPE
from tempfile import TemporaryDirectory
from unittest import TestCase

from jupyter_client.kernelspec import NATIVE_KERNEL_NAME
import pytest
from tornado.testing import AsyncTestCase, gen_test
from traitlets.config.loader import Config

from ..client_helper import ExecClient

try:
    from .. import (
        PooledKernelManager,
        PooledMappingKernelManager,
        MaximumKernelsException,
    )
except ImportError:
    pass

from .utils import async_shutdown_all_direct


class TestUpdatePooled(AsyncTestCase):
    @gen_test
    async def test_kernel_ok(self):
        # Just test that we can start a kernel and communicate with it.
        c = Config()
        c.PooledKernelManager.kernel_pools = {NATIVE_KERNEL_NAME: 1}
        c.PooledKernelManager.strict_pool_names = True
        km = PooledKernelManager(config=c)

        kid = await km.start_kernel()
        kernel = km.get_kernel(kid)
        self.assertIsNotNone(kernel)

        try:
            client = ExecClient(kernel, _store_outputs=True)
            async with client.setup_kernel():
                await client.execute('import os\nprint(os.environ.get("VARFOO"))')
            self.assertEqual(
                client._outputs,
                [{"name": "stdout", "output_type": "stream", "text": "None\n"}],
            )
        finally:
            await km.shutdown_all()

    @gen_test
    async def test_env(self):
        c = Config()
        c.PooledKernelManager.kernel_pools = {NATIVE_KERNEL_NAME: 1}
        c.PooledKernelManager.strict_pool_names = True
        km = PooledKernelManager(config=c)

        try:
            kid = await km.start_kernel(env={"VARFOO": "BARBAR"})
            kernel = km.get_kernel(kid)
            self.assertIsNotNone(kernel)

            client = ExecClient(kernel, _store_outputs=True)
            async with client.setup_kernel():
                await client.execute('import os\nprint(os.environ.get("VARFOO"))')
            self.assertEqual(
                client._outputs,
                [{"name": "stdout", "output_type": "stream", "text": "BARBAR\n"}],
            )
        finally:
            await km.shutdown_all()

    @gen_test
    async def test_cwd(self):
        with TemporaryDirectory() as tmp_dir:
            c = Config()
            c.PooledKernelManager.kernel_pools = {NATIVE_KERNEL_NAME: 1}
            c.PooledKernelManager.strict_pool_names = True
            km = PooledKernelManager(config=c)

            # create a local py module in cwd to test import
            foo_mod = Path(tmp_dir) / "foo_module.py"
            foo_mod.write_text("foo = 1")

            try:
                kid = await km.start_kernel(cwd=tmp_dir)
                kernel = km.get_kernel(kid)
                self.assertIsNotNone(kernel)

                client = ExecClient(kernel, _store_outputs=True)
                async with client.setup_kernel():
                    await client.execute(
                        "import foo_module, pathlib\nprint(pathlib.Path(foo_module.__file__).resolve())"
                    )
                self.assertEqual(
                    client._outputs,
                    [
                        {
                            "name": "stdout",
                            "output_type": "stream",
                            "text": f"{foo_mod.resolve()}\n",
                        }
                    ],
                )
            finally:
                await km.shutdown_all()



from contextlib import contextmanager
from subprocess import PIPE

from tornado.testing import gen_test
from traitlets.config.loader import Config

from .. import (
    SyncLimitedKernelManager,
    MaximumKernelsException,
)

from .utils_sync import TestKernelManager


# Test that it works as normal with default config
class TestLimitedKernelManager(TestKernelManager):
    __test__ = True

    # static so picklable for multiprocessing on Windows
    @staticmethod
    @contextmanager
    def _get_tcp_km():
        c = Config()
        km = SyncLimitedKernelManager(config=c)
        try:
            yield km
        finally:
            km.shutdown_all(now=True)


# Test that it works with a max of 4
class TestLimitedKernelManagerApplied(TestKernelManager):
    __test__ = True

    # static so picklable for multiprocessing on Windows
    @staticmethod
    @contextmanager
    def _get_tcp_km():
        c = Config()
        c.SyncLimitedKernelManager.max_kernels = 4
        km = SyncLimitedKernelManager(config=c)
        try:
            yield km
        finally:
            km.shutdown_all(now=True)

    def test_touch_max(self):
        with self._get_tcp_km() as km:
            kids = []
            for i in range(4):
                kid = km.start_kernel(stdout=PIPE, stderr=PIPE)
                self.assertIn(kid, km)
                kids.append(kid)

            # Remove and add one to make sure we don't count closed kernels
            km.shutdown_kernel(kid)
            self.assertNotIn(kid, km)
            kids.pop()

            kid = km.start_kernel(stdout=PIPE, stderr=PIPE)
            self.assertIn(kid, km)
            kids.append(kid)

            km.shutdown_all()
            for kid in kids:
                self.assertNotIn(kid, km)

    def test_breach_max(self):
        with self._get_tcp_km() as km:
            kids = []
            for i in range(4):
                kid = km.start_kernel(stdout=PIPE, stderr=PIPE)
                self.assertIn(kid, km)
                kids.append(kid)

            with self.assertRaises(MaximumKernelsException):
                km.start_kernel(stdout=PIPE, stderr=PIPE)

            # Remove and add one to make sure we correctly recovered
            km.shutdown_kernel(kid)
            self.assertNotIn(kid, km)
            kids.pop()

            kid = km.start_kernel(stdout=PIPE, stderr=PIPE)
            self.assertIn(kid, km)
            kids.append(kid)

            km.shutdown_all()
            for kid in kids:
                self.assertNotIn(kid, km)



# coding: utf-8

# Copyright (c) Vidar Tonaas Fauske.
# Distributed under the terms of the Modified BSD License.
"""Hotpot - Jupyter kernel manager helpers

This module contains
"""

import asyncio

from jupyter_client.multikernelmanager import MultiKernelManager
from traitlets import Bool, Dict, Float, Integer, List, Unicode, observe

from .async_utils import ensure_event_loop, just_run
from .client_helper import ExecClient
from .limited import SyncLimitedKernelManager, MaximumKernelsException
from .py_snippets import (
    python_update_cwd_code,
    python_update_env_code,
    python_init_import_code,
)


async def _wait_before(delay, aw):
    await asyncio.sleep(delay)
    return await aw


def _ensure_event_loop():
    try:
        loop = asyncio.get_event_loop()
    except RuntimeError:
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
    return loop


class SyncPooledKernelManager(SyncLimitedKernelManager):
    kernel_pools = Dict(
        Integer(0),
        config=True,
        help="Mapping from kernel name to the number of started kernels to keep on standby",
    )

    pool_kwargs = Dict(
        Dict(),
        config=True,
        help="Mapping from kernel name to the arguments passed to kernel_start when pre-warming",
    )

    strict_pool_names = Bool(
        config=True,
        help="Whether to allow starting kernels with other names than those explicitly listed in kernel_pools",
    )

    strict_pool_kwargs = Bool(
        config=True,
        help="Whether to allow starting kernels with other kwargs than those explicitly listed in pool_kwargs",
    )

    fill_delay = Float(
        1,
        config=True,
        help="Wait time before re-filling the pool after a kernel is used",
    )

    initialization_code = Dict(config=True, help="Code that gets executed at startup")

    python_imports = List(
        Unicode(), [], config=True, help="List of Python modules/packages to import"
    )

    _pools = Dict()
    _init_futs = Dict()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fill_if_needed(delay=0)
        self.observe(self._pool_size_changed, "kernel_pools")
        self._discarded = []

    def _pool_size_changed(self, change):
        self.unfill_as_needed()
        self.fill_if_needed()

    def _should_use_pool(self, kernel_name, kwargs):
        """Verify name and kwargs, and check whether we should use the pool"""
        if "kernel_id" in kwargs:
            return False

        if self.strict_pool_names and kernel_name not in self.kernel_pools:
            raise ValueError("Cannot start kernel with name %r" % (kernel_name,))
        if self.strict_pool_kwargs and (
            kernel_name not in self.pool_kwargs or kwargs != self.pool_kwargs[kernel_name]
        ):
            raise ValueError("Cannot start kernel with kwargs %r" % (kwargs,))

        return len(self._pools.get(kernel_name, ())) > 0

    def unfill_as_needed(self):
        """Kills extra kernels in pool"""
        tasks = []
        for name, target in self.kernel_pools.items():
            pool = self._pools.get(name, [])
            self._pools[name] = pool
            for i in range(len(pool) - target):
                kernel_id = pool.pop(0)
                self.shutdown_kernel(kernel_id)

    def fill_if_needed(self, delay=None):
        """Start kernels until pool is full"""
        loop = ensure_event_loop()
        for name, target in self.kernel_pools.items():
            pool = self._pools.get(name, [])
            self._pools[name] = pool
            for i in range(target - len(pool)):
                kw = self.pool_kwargs.get(name, {})
                kernel_id = just_run(super().start_kernel(kernel_name=name, **kw))
                # Todo: use delay
                # Start the work on the loop immediately, so it is ready when needed:
                self._init_futs[kernel_id] = loop.create_task(self._initialize(name, kernel_id))
                pool.append(kernel_id)

    async def wait_for_pool(self):
        await asyncio.gather(*self._init_futs.values())

    async def _pop_pooled_kernel(self, kernel_name, kwargs):
        self.log.debug("Using kernel from pool: %s", kernel_name)
        kernel_id = self._pools[kernel_name].pop(0)
        await self._init_futs.pop(kernel_id)
        return await self._update_kernel(kernel_name, kernel_id, kwargs)

    def start_kernel(self, kernel_name=None, **kwargs):
        if kernel_name is None:
            kernel_name = self.default_kernel_name
        self.log.debug("Starting kernel: %s", kernel_name)
        kernel_id = kwargs.get("kernel_id")
        if kernel_id is None and self._should_use_pool(kernel_name, kwargs):
            kernel_id = just_run(self._pop_pooled_kernel(kernel_name, kwargs))
        else:
            kernel_id = just_run(super().start_kernel(kernel_name=kernel_name, **kwargs))

        try:
            self.fill_if_needed()
        except MaximumKernelsException:
            pass
        return kernel_id

    def restart_kernel(self, kernel_id, **kwargs):
        km = self.get_kernel(kernel_id)
        kernel_name = km.kernel_name
        just_run(super().restart_kernel(kernel_id, **kwargs))
        just_run(self._update_kernel(kernel_name, kernel_id, km._launch_args))
        just_run(self._initialize(kernel_name, kernel_id))

    def shutdown_kernel(self, kernel_id, *args, **kwargs):
        if kernel_id in self._init_futs:
            self._init_futs.pop(kernel_id).cancel()
        return super().shutdown_kernel(kernel_id, *args, **kwargs)

    def shutdown_all(self, *args, **kwargs):
        for pool in self._pools.values():
            # The iteration gets confused if we don't copy pool
            for kernel_id in tuple(pool):
                self.shutdown_kernel(kernel_id, *args, **kwargs)
        for kernel_id in self.list_kernel_ids():
            self.shutdown_kernel(kernel_id, *args, **kwargs)
        self._pools = {}
        self._init_futs = {}

    async def _update_kernel(self, kernel_name, kernel_id, kwargs):
        base_kws = self.pool_kwargs.get(kernel_name)
        if base_kws:
            new_kws = {}
            for k, v in kwargs.items():
                if k not in base_kws:
                    new_kws[k] = v
                elif base_kws[k] != kwargs[k]:
                    new_kws[k] = v
        else:
            new_kws = kwargs

        # Make sure that the kernel is in a state that matches kwargs
        # Currently supported is a python kernel, and the path/cwd and env arguments
        if kernel_name in ("python3", "python") and new_kws:
            # Avoid client overhead if not needed:
            if "path" in new_kws or "cwd" in new_kws or "env" in new_kws:
                kernel = self.get_kernel(kernel_id)
                client = ExecClient(kernel)
                async with client.setup_kernel():
                    if "path" in new_kws:
                        new_kws["cwd"] = self.cwd_for_path(new_kws.pop("path"))
                    if "cwd" in new_kws:
                        cwd = new_kws.pop("cwd")
                        code = python_update_cwd_code.format(cwd=cwd)
                        self.log.debug("Updating preheated kernel CWD using")
                        await client.execute(code)
                    if "env" in new_kws:
                        env = new_kws.pop("env")
                        code = python_update_env_code.format(env=env)
                        self.log.debug("Updating preheated kernel env vars")
                        await client.execute(code)
        if new_kws:
            self.log.debug("Unknown kwargs: %s", list(new_kws.keys()))

        return kernel_id

    async def _initialize(self, kernel_name, kernel_id):
        """Run any configured initialization code in the kernel"""
        extension = None
        language = None

        kernel = self.get_kernel(kernel_id)

        try:
            language_to_extensions = {"python": "py"}
            language = kernel.kernel_spec_manager.get_all_specs()[kernel_name]["spec"]["language"]
            extension = language_to_extensions[language]
        except Exception:
            pass

        py_imports = language == "python" and self.python_imports

        if not extension and not py_imports:
            # Save some effort
            return kernel_id

        self.log.info("Initializing kernel: %s", kernel_id)

        client = ExecClient(kernel)

        from jupyter_core.paths import jupyter_config_path
        from pathlib import Path

        async with client.setup_kernel():
            if extension:
                for base_path in map(Path, jupyter_config_path()):
                    path = base_path / f"kernel_pool_init_{kernel_name}.{extension}"
                    if path.exists():
                        with open(path) as f:
                            self.log.debug("Running %s for initializing kernel", path)
                            code = f.read()
                        await client.execute(code)
            if py_imports:
                code = python_init_import_code.format(modules=self.python_imports)
                await client.execute(code)
        self.log.info("Initialized kernel: %s", kernel_id)
        return kernel_id


__all__ = [
    "SyncPooledKernelManager",
]
